from ai_core_sdk.ai_core_v2_client import AICoreV2Client
from hdbcli import dbapi

from .config.config import Config
from .aicore.aicore import AICore
from .hanadb.hanadb import HANADb
from .exceptions.exceptions import GenAIWrapperException
from .data_objects.chat_object import ChatObject
from .data_objects.vector_object import HANAVectorObject
from .comm.chat import Chat 
from .comm.embed import Embeddings


central_config: dict = None
ai_core_client: AICoreV2Client = None
hana_db_client: dbapi = None

__all__ = [
    "central_config",    
    "ai_core_client",

    "GenAIWrapperException",

    "Config",
    "AICore",
    "HANADb",
    "Chat",
    "Embeddings",
    "ChatObject",
    "HANAVectorObject"
]


import requests
import json

import genai_wrapper
from genai_wrapper import Config, GenAIWrapperException

class BaseComm:
    def __init__( self, **kwargs: dict ) -> None:
        self.bu = genai_wrapper.central_config["ai_core"]["secret"]["serviceurls"]["AI_API_URL"]
        self.cu = kwargs["cu"]
        self.dp = kwargs["dp"]
        self.bt = genai_wrapper.ai_core_client.rest_client.get_token()
        self.rg = genai_wrapper.central_config["ai_core"]["resource_group"]
        self.pm = kwargs["pm"]

    def query(self: object, data: dict) -> dict:
        result = requests.post(
            f"{self.bu}/v2/inference/deployments/{self.dp}/{self.cu}",
            headers={k:v for k,v in zip(Config.AI.AI_REQH, [self.bt, self.rg, "application/json"])}, 
            data=json.dumps(data | self.pm)
        )
        if result.status_code == 200:
            return json.loads( result.content )
        else:
            raise GenAIWrapperException(f"http request error (code: {result.status_code}) -> {result.text}")



import json

from genai_wrapper import GenAIWrapperException

class ChatObject:
    def __init__(self) -> None:
        self.messages = [
            {
                "role": "user",
                "content": ""
            }
        ]

    def _get_message_(self) -> list:
        return self.messages
    
    def _process_output_(self, output: str) -> None:
        self._response_ = output
        try:
            self._message: dict = output["choices"][0]["message"]
            self._usage: dict = output["usage"]
        except:
            raise GenAIWrapperException(f"Message does not contain the correct response.\n{self._response_}")
        
        return self

    def add_role(self, role: str, content: str, append: bool=False) -> None:
        if any(map(lambda message: message['role'] == 'user', self.messages)):
            for message in self.messages:
                if message['role'] == 'user':
                    message['content'] = content if not append else message['content'] + content
        else:
            self.messages.append({"role": role, "content": content})

    def add_message(self, content: str, append: bool=False) -> None:
        self.add_role("user", content, append)

    def get_answer(self) -> str:
        try:
            return self._message["content"]
        except:
            raise GenAIWrapperException(f"No query has been made to llm or the response was malformed!")

    def get_usage(self) -> dict:
        try:
            return self._usage
        except:
            raise GenAIWrapperException(f"No query has been made to llm or the response was malformed!")



import json

import genai_wrapper
from genai_wrapper import (
    Config,
    ChatObject
)
from .base_comm import BaseComm

class Chat(BaseComm):
    def __init__(self, llm_model: str) -> None:
        self.llm_model = llm_model
        super().__init__(
            cu=f"{Config.AI.AI_URLTYPE[0]}?{Config.AI.AI_VER}",
            dp=genai_wrapper.central_config["gen_ai"][llm_model]["deploymentid"],
            pm=Config.AI.AI_PARAMS | genai_wrapper.central_config["gen_ai"][llm_model]["parameters"]
        )

    def query(self, chat_object: ChatObject) -> ChatObject:
        payload = { "messages": chat_object._get_message_() }
        if not genai_wrapper.central_config["gen_ai"][self.llm_model]["model_name"].lower().startswith("gpt"):
            payload = { "model": genai_wrapper.central_config["gen_ai"][self.llm_model]["model_name"], "stream": False  } | payload        
        return chat_object._process_output_(
            super().query( payload )
        )


import json

from genai_wrapper import GenAIWrapperException

class HANAVectorObject:
    def __init__( self, **kwargs: dict ) -> None:
        self.table = kwargs.get( "table", None )
        self.columns = kwargs.get( "columns", None )
        if not self.table:
            raise GenAIWrapperException("Db table cannot be empty!")
        else:
            if type(self.columns) == list:
                self.columns = ", ".join(self.columns)
        self.vector_col = kwargs.get( "vector_col", None )
        if not self.vector_col:
            raise GenAIWrapperException("You need to pass at least 1 vector store column")
        self.conditions = kwargs.get( "conditions", None )
        if self.conditions:
            self.conditions = None if len( self.conditions.trim() ) == 0 else self.conditions
        self.k = kwargs.get( "k", 3 )

    def _process_output_( self, records: list ) -> dict:
        self._result = {
            "total_records": 0,
            "records": {}
        }
        if len( records ) > 0:
            self._result = {
                "total_records": len( records ),
                "records": { key.strip(): [] for key in records[0].column_names }
            }
            for row in records:
                for col in self._result["records"].keys():
                    self._result["records"][col].append( row[col] )

        return self._result


import genai_wrapper
from genai_wrapper import AICoreV2Client

class AICore:
    def init() -> None:
        genai_wrapper.ai_core_client = AICoreV2Client(
            base_url=genai_wrapper.central_config["ai_core"]["secret"]["serviceurls"]["AI_API_URL"],
            auth_url=genai_wrapper.central_config["ai_core"]["secret"]["url"]  + ( "/oauth/token" if "/oauth/token" not in genai_wrapper.central_config["ai_core"]["secret"]["serviceurls"]["AI_API_URL"] else "" ),
            client_id=genai_wrapper.central_config["ai_core"]["secret"]["clientid"],
            client_secret=genai_wrapper.central_config["ai_core"]["secret"]["clientsecret"]
        )

    def get_client() -> AICoreV2Client:
        return genai_wrapper.ai_core_client




import genai_wrapper
from genai_wrapper import dbapi

class HANADb:
    def init() -> None:
        genai_wrapper.hana_db_client = dbapi.connect(
            address=genai_wrapper.central_config["hana_vec_store"]["host"],
            port=genai_wrapper.central_config["hana_vec_store"]["port"],
            user=genai_wrapper.central_config["hana_vec_store"]["userid"],
            password=genai_wrapper.central_config["hana_vec_store"]["password"],
            sslValidateCertificate=genai_wrapper.central_config["hana_vec_store"]["ssl_cert_validation"]
        )

    def get_client() -> dbapi:
        return genai_wrapper.hana_db_client
    
    def close() -> None:
        if genai_wrapper.hana_db_client:
            genai_wrapper.hana_db_client.close()




# **********************************************************************
# Config file, required by the package.
# *DO NOT MODIFY THIS FILE*
# **********************************************************************
class Config:
    class AI:
        AI_REQH: list = [
            "Authorization",
            "AI-Resource-Group",
            "Content-Type"
        ]
        AI_URLTYPE: list = [
            "chat/completions",
            "embeddings"
        ]
        AI_VER: str = "api-version=2023-05-15"
        AI_PARAMS: dict = {
            "max_tokens": 256,
            "temperature": 0.0,
            "frequency_penalty": 0,
            "presence_penalty": 0,
            "stop": "null"
        }


from setuptools import setup, find_packages
from pathlib import Path

current_dir = Path( __file__ ).parent
readme = (current_dir / "README.md" ).read_text()

setup(
    name="genai_wrapper",
    version="0.0.8",
    packages=find_packages(),
    install_requires=[
        "ai-api-client-sdk>=2.1.4",
        "ai-core-sdk>=2.3.11",
        "hdbcli>=2.21.28"
    ],
    author="Praveen Nair",
    author_email="",
    description="A user-friendly GenAI wrapper that leverages SAP's AI Core to seamlessly translate requests into LLM calls.",
    long_description=readme,
    long_description_content_type="text/markdown",
    license="MIT",
    project_urls={
        "Source Repository": "https://github.com/praveen-nair/genai_wrapper",
        "Issues": "https://github.com/praveen-nair/genai_wrapper/issues"
    }
)



import genai_wrapper
from genai_wrapper import Config
from .base_comm import BaseComm

class Embeddings(BaseComm):
    def __init__( self, llm_model: str ) -> None:
        super().__init__(
            cu=f"{Config.AI.AI_URLTYPE[1]}?{Config.AI.AI_VER}",
            dp=genai_wrapper.central_config["gen_ai"][llm_model]["deploymentid"],
            pm={}
        )

    def query( self, text ):
        return super().query( { "input": text } )["data"][0]["embedding"]


import json

import genai_wrapper
from genai_wrapper import *

class GenAIWrapper:
    __author__ = "https://github.com/praveen-nair"
    __version__ = "00.00.08.202411"

    def __init__( self, **kwargs: dict ) -> None:
        global central_config
        config_content = kwargs.get("config_file", None)
        if config_content:
            with open( config_content, 'r' ) as file:
                config_content = json.load( file )
        else:
            config_content = kwargs.get("config_content", None)
            if config_content:
                config_content = json.loads( config_content )
            else:
                config_content = kwargs.get("config_dict", None)
                if not config_content:
                    raise GenAIWrapperException("Either config_file, config_content, or config_dict must be provided as input.")

        genai_wrapper.central_config = config_content
        AICore.init()

    def chat( self, llm_model: dict, chat_object: ChatObject ) -> str:
        return Chat(
            llm_model=llm_model
        ).query(chat_object)
        
    def embedding( self, llm_model: dict, text: str ) -> list:
        return Embeddings(
            llm_model=llm_model
        ).query( text=text )
    
    def embedding_vec_store( self, llm_model: dict, hana_vector_object: HANAVectorObject, **kwargs: dict ) -> None:
        if not genai_wrapper.hana_db_client:
            HANADb.init()

        hana_vector_object.k = kwargs.get( "k", hana_vector_object.k )
        vec_query = kwargs.get("vec_value", None)
        if not vec_query:
            vec_query = kwargs.get("vec_text", None)
            if not vec_query:
                raise GenAIWrapperException("Either vec_value or vec_text must be passed!")
            else:
                vec_query = self.embedding( llm_model, vec_query )

        cursor = genai_wrapper.hana_db_client.cursor()

        cursor.execute( 
            f"SELECT TOP {hana_vector_object.k} {hana_vector_object.columns}, " f"COSINE_SIMILARITY({hana_vector_object.vector_col}, TO_REAL_VECTOR('{vec_query}')) AS SIM_SCORE " f"FROM {hana_vector_object.table} " f"{'WHERE ' + hana_vector_object.conditions if hana_vector_object.conditions else ''} " f"ORDER BY SIM_SCORE DESC;" 
        )

        result = hana_vector_object._process_output_( cursor.fetchall() )

        cursor.close()

        return result

    def close( self ):
        if not genai_wrapper.hana_db_client:
            HANADb.close() 